# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Reload:
    """

    Attributes
    ----------
    appId: str
      The ID of the app.
    creationTime: str
      The time the reload job was created.
    endTime: str
      The time the reload job finished.
    engineTime: str
      The timestamp returned from the Sense engine upon successful reload.
    id: str
      The ID of the reload.
    links: ReloadLinks
    log: str
      The log describing the result of the reload request.
    partial: bool
      The boolean value used to present the reload is partial or not
    startTime: str
      The time the reload job was consumed from the queue.
    status: Literal["QUEUED", "RELOADING", "CANCELING", "SUCCEEDED", "FAILED", "CANCELED", "EXCEEDED_LIMIT"]
      The status of the reload. There are seven statuses. `QUEUED`, `RELOADING`, `CANCELING` are the active statuses. `SUCCEEDED`, `FAILED`,`CANCELED`,`EXCEEDED_LIMIT` are the end statuses.
    tenantId: str
      The ID of the tenant who owns the reload.
    type: str
      The type of reload event (hub, chronos or external).
    userId: str
      The ID of the user who created the reload.
    """

    appId: str = None
    creationTime: str = None
    endTime: str = None
    engineTime: str = None
    id: str = None
    links: ReloadLinks = None
    log: str = None
    partial: bool = None
    startTime: str = None
    status: Literal[
        "QUEUED",
        "RELOADING",
        "CANCELING",
        "SUCCEEDED",
        "FAILED",
        "CANCELED",
        "EXCEEDED_LIMIT",
    ] = None
    tenantId: str = None
    type: str = None
    userId: str = None

    def __init__(self_, **kvargs):

        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "creationTime" in kvargs and kvargs["creationTime"] is not None:
            self_.creationTime = kvargs["creationTime"]
        if "endTime" in kvargs and kvargs["endTime"] is not None:
            self_.endTime = kvargs["endTime"]
        if "engineTime" in kvargs and kvargs["engineTime"] is not None:
            self_.engineTime = kvargs["engineTime"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Reload.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = ReloadLinks(**kvargs["links"])
        if "log" in kvargs and kvargs["log"] is not None:
            self_.log = kvargs["log"]
        if "partial" in kvargs and kvargs["partial"] is not None:
            self_.partial = kvargs["partial"]
        if "startTime" in kvargs and kvargs["startTime"] is not None:
            self_.startTime = kvargs["startTime"]
        if "status" in kvargs and kvargs["status"] is not None:
            self_.status = kvargs["status"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "userId" in kvargs and kvargs["userId"] is not None:
            self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def cancel(self) -> None:
        """
        Cancels a reload
        Cancels a reload that is in progress or has been queued

        Parameters
        ----------
        """
        self.auth.rest(
            path="/reloads/{reloadId}/actions/cancel".replace("{reloadId}", self.id),
            method="POST",
            params={},
            data=None,
        )


@dataclass
class Href:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ReloadLinks:
    """

    Attributes
    ----------
    self: Href
    """

    self: Href = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == ReloadLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ReloadRequest:
    """

    Attributes
    ----------
    appId: str
      The ID of the app to be reloaded.
    partial: bool
      The boolean value used to present the reload is partial or not
    """

    appId: str = None
    partial: bool = None

    def __init__(self_, **kvargs):

        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "partial" in kvargs and kvargs["partial"] is not None:
            self_.partial = kvargs["partial"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ReloadsClass:
    """

    Attributes
    ----------
    data: list[Reload]
    links: ReloadsLinks
    """

    data: list[Reload] = None
    links: ReloadsLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == ReloadsClass.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Reload(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == ReloadsClass.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = ReloadsLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ReloadsLinks:
    """

    Attributes
    ----------
    self: Href
    next: Href
    prev: Href
    """

    self: Href = None
    next: Href = None
    prev: Href = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == ReloadsLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == ReloadsLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Href(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == ReloadsLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Href(**kvargs["prev"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Reloads:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get(self, reloadId: str) -> Reload:
        """
        Get reload record
        Finds and returns a reload record

        Parameters
        ----------
        reloadId: str
          The unique identifier of the reload.
        """
        response = self.auth.rest(
            path="/reloads/{reloadId}".replace("{reloadId}", reloadId),
            method="GET",
            params={},
            data=None,
        )
        obj = Reload(**response.json())
        obj.auth = self.auth
        return obj

    def get_reloads(
        self,
        appId: str,
        limit: int = 10,
        next: str = None,
        partial: bool = None,
        prev: str = None,
        max_items: int = 10,
    ) -> ListableResource[Reload]:
        """
        Finds and returns the reloads that the user has access to.

        Parameters
        ----------
        appId: str
          The UUID formatted string used to search for an app's reload history entries. TenantAdmin users may omit this parameter to list all reload history in the tenant.
        limit: int = 10
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        partial: bool = None
          The boolean value used to search for a reload is partial or not.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        """
        query_params = {}
        if appId is not None:
            query_params["appId"] = appId
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if partial is not None:
            query_params["partial"] = partial
        if prev is not None:
            query_params["prev"] = prev
        response = self.auth.rest(
            path="/reloads",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Reload,
            auth=self.auth,
            path="/reloads",
            max_items=max_items,
            query_params=query_params,
        )

    def create(self, data: ReloadRequest) -> Reload:
        """
        Reloads an app specified by an app ID.

        Parameters
        ----------
        data: ReloadRequest
          Request body specifying ID of app to be reloaded.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/reloads",
            method="POST",
            params={},
            data=data,
        )
        obj = Reload(**response.json())
        obj.auth = self.auth
        return obj



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class CollectionResultResponseBody:
    """
    A collection.

    Attributes
    ----------
    createdAt: str
      The RFC3339 datetime when the collection was created.
    creatorId: str
      The ID of the user who created the collection. This property is only populated if the JWT contains a userId.
    description: str
    id: str
      The collection's unique identifier.
    itemCount: int
      The number of items that have been added to the collection.
    links: CollectionLinksResponseBody
    meta: CollectionMetaResponseBody
      Collection metadata and computed fields.
    name: str
    tenantId: str
      The ID of the tenant that owns the collection. This property is populated by using JWT.
    type: Literal["public", "private", "favorite"]
      The collection's type.
    updatedAt: str
      The RFC3339 datetime when the collection was last updated.
    updaterId: str
      The ID of the user who last updated the collection. This property is only populated if the JWT contains a userId.
    """

    createdAt: str = None
    creatorId: str = None
    description: str = None
    id: str = None
    itemCount: int = None
    links: CollectionLinksResponseBody = None
    meta: CollectionMetaResponseBody = None
    name: str = None
    tenantId: str = None
    type: Literal["public", "private", "favorite"] = None
    updatedAt: str = None
    updaterId: str = None

    def __init__(self_, **kvargs):

        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "creatorId" in kvargs and kvargs["creatorId"] is not None:
            self_.creatorId = kvargs["creatorId"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "itemCount" in kvargs and kvargs["itemCount"] is not None:
            self_.itemCount = kvargs["itemCount"]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == CollectionResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = CollectionLinksResponseBody(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if (
                type(kvargs["meta"]).__name__
                == CollectionResultResponseBody.__annotations__["meta"]
            ):
                self_.meta = kvargs["meta"]
            else:
                self_.meta = CollectionMetaResponseBody(**kvargs["meta"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "updaterId" in kvargs and kvargs["updaterId"] is not None:
            self_.updaterId = kvargs["updaterId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete_item(self, itemId: str) -> None:
        """
        Removes an item from a collection.
        Removes an item from a collection.


        Parameters
        ----------
        itemId: str
          The item's unique identifier.
        """
        self.auth.rest(
            path="/collections/{collectionId}/items/{itemId}".replace(
                "{collectionId}", self.id
            ).replace("{itemId}", itemId),
            method="DELETE",
            params={},
            data=None,
        )

    def get_item(self, itemId: str) -> ItemResultResponseBody:
        """
        Returns an item in a specific collection.
        Finds and returns an item. See GET /items/{id}


        Parameters
        ----------
        itemId: str
          The item's unique identifier.
        """
        response = self.auth.rest(
            path="/collections/{collectionId}/items/{itemId}".replace(
                "{collectionId}", self.id
            ).replace("{itemId}", itemId),
            method="GET",
            params={},
            data=None,
        )
        obj = ItemResultResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def get_items(
        self,
        limit: int = 10,
        name: str = None,
        next: str = None,
        prev: str = None,
        query: str = None,
        resourceId: str = None,
        resourceLink: str = None,
        resourceType: Literal[
            "app",
            "collection",
            "qlikview",
            "insight",
            "qvapp",
            "genericlink",
            "sharingservicetask",
            "note",
            "dataasset",
            "dataset",
            "automation",
            "automl-experiment",
            "automl-deployment",
        ] = None,
        shared: bool = None,
        sort: Literal[
            "+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"
        ] = None,
        spaceId: str = None,
        noActions: bool = False,
    ) -> CollectionsListCollectionItemsResponseBody:
        """
        Retrieves items in a collection.
        Finds and returns items from a collection that the user has access to.


        Parameters
        ----------
        limit: int = 10
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        name: str = None
          The case-insensitive string used to search for a resource by name.
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        query: str = None
          The case-insensitive string used to search for a resource by name or description.
        resourceId: str = None
          The case-sensitive string used to search for an item by resourceId. If resourceId is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
        resourceLink: str = None
          The case-sensitive string used to search for an item by resourceLink. If resourceLink is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
        resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"] = None
          The case-sensitive string used to search for an item by resourceType.
        shared: bool = None
          Whether or not to return items in a shared space.
        sort: Literal["+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"] = None
          The property of a resource to sort on (default sort is +createdAt). The supported properties are createdAt, updatedAt, and name. A property must be prefixed by + or   - to indicate ascending or descending sort order respectively.
        spaceId: str = None
          The space's unique identifier (supports \'personal\' as spaceId).
        noActions: bool = False
          If set to true, the user's available actions for each item will not be evaluated meaning the actions-array will be omitted from the response (reduces response time).

        """
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if query is not None:
            query_params["query"] = query
        if resourceId is not None:
            query_params["resourceId"] = resourceId
        if resourceLink is not None:
            query_params["resourceLink"] = resourceLink
        if resourceType is not None:
            query_params["resourceType"] = resourceType
        if shared is not None:
            query_params["shared"] = shared
        if sort is not None:
            query_params["sort"] = sort
        if spaceId is not None:
            query_params["spaceId"] = spaceId
        if noActions is not None:
            query_params["noActions"] = noActions
        response = self.auth.rest(
            path="/collections/{collectionId}/items".replace("{collectionId}", self.id),
            method="GET",
            params=query_params,
            data=None,
        )
        obj = CollectionsListCollectionItemsResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def create_item(
        self, data: CollectionsAddCollectionItemRequestBody
    ) -> ItemResultResponseBody:
        """
        Adds an item to a collection.
        Adds an item to a collection and returns the item.


        Parameters
        ----------
        data: CollectionsAddCollectionItemRequestBody

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/collections/{collectionId}/items".replace("{collectionId}", self.id),
            method="POST",
            params={},
            data=data,
        )
        obj = ItemResultResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def delete(self) -> None:
        """
        Deletes a collection.
        Deletes a collection and removes all items from the collection.


        Parameters
        ----------
        """
        self.auth.rest(
            path="/collections/{collectionId}".replace("{collectionId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def set(
        self, data: CollectionsUpdateCollectionRequestBody
    ) -> CollectionResultResponseBody:
        """
        Updates a collection.
        Updates a collection and returns the new collection. Omitted and unsupported fields are ignored. To unset a field, provide the field's zero value.


        Parameters
        ----------
        data: CollectionsUpdateCollectionRequestBody

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/collections/{collectionId}".replace("{collectionId}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self


@dataclass
class CollectionsAddCollectionItemRequestBody:
    """

    Attributes
    ----------
    id: str
      The item's unique identifier.
    """

    id: str = None

    def __init__(self_, **kvargs):

        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsCreateCollectionRequestBody:
    """

    Attributes
    ----------
    description: str
    name: str
    type: Literal["public", "private"]
      The collection's type.
    """

    description: str = None
    name: str = None
    type: Literal["public", "private"] = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsListCollectionItemsResponseBody:
    """
    ListCollectionItemsResponseBody result type

    Attributes
    ----------
    data: list[ItemResultResponseBody]
    """

    data: list[ItemResultResponseBody] = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == CollectionsListCollectionItemsResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [ItemResultResponseBody(**e) for e in kvargs["data"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsListCollectionsResponseBody:
    """
    ListCollectionsResponseBody result type

    Attributes
    ----------
    data: list[CollectionResultResponseBody]
    links: CollectionsLinksResponseBody
    """

    data: list[CollectionResultResponseBody] = None
    links: CollectionsLinksResponseBody = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == CollectionsListCollectionsResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [CollectionResultResponseBody(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == CollectionsListCollectionsResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = CollectionsLinksResponseBody(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsUpdateCollectionRequestBody:
    """

    Attributes
    ----------
    description: str
    name: str
    """

    description: str = None
    name: str = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionLinksResponseBody:
    """

    Attributes
    ----------
    items: Link
    self: Link
    """

    items: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "items" in kvargs and kvargs["items"] is not None:
            if (
                type(kvargs["items"]).__name__
                == CollectionLinksResponseBody.__annotations__["items"]
            ):
                self_.items = kvargs["items"]
            else:
                self_.items = Link(**kvargs["items"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == CollectionLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionMetaResponseBody:
    """
    Collection metadata and computed fields.

    Attributes
    ----------
    items: ItemsResultResponseBody
      Multiple items.
    """

    items: ItemsResultResponseBody = None

    def __init__(self_, **kvargs):

        if "items" in kvargs and kvargs["items"] is not None:
            if (
                type(kvargs["items"]).__name__
                == CollectionMetaResponseBody.__annotations__["items"]
            ):
                self_.items = kvargs["items"]
            else:
                self_.items = ItemsResultResponseBody(**kvargs["items"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsLinksResponseBody:
    """

    Attributes
    ----------
    item: Link
    next: Link
    prev: Link
    self: Link
    """

    item: Link = None
    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "item" in kvargs and kvargs["item"] is not None:
            if (
                type(kvargs["item"]).__name__
                == CollectionsLinksResponseBody.__annotations__["item"]
            ):
                self_.item = kvargs["item"]
            else:
                self_.item = Link(**kvargs["item"])
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == CollectionsLinksResponseBody.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == CollectionsLinksResponseBody.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == CollectionsLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemLinksResponseBody:
    """

    Attributes
    ----------
    collections: Link
    open: Link
    self: Link
    thumbnail: Link
    """

    collections: Link = None
    open: Link = None
    self: Link = None
    thumbnail: Link = None

    def __init__(self_, **kvargs):

        if "collections" in kvargs and kvargs["collections"] is not None:
            if (
                type(kvargs["collections"]).__name__
                == ItemLinksResponseBody.__annotations__["collections"]
            ):
                self_.collections = kvargs["collections"]
            else:
                self_.collections = Link(**kvargs["collections"])
        if "open" in kvargs and kvargs["open"] is not None:
            if (
                type(kvargs["open"]).__name__
                == ItemLinksResponseBody.__annotations__["open"]
            ):
                self_.open = kvargs["open"]
            else:
                self_.open = Link(**kvargs["open"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == ItemLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        if "thumbnail" in kvargs and kvargs["thumbnail"] is not None:
            if (
                type(kvargs["thumbnail"]).__name__
                == ItemLinksResponseBody.__annotations__["thumbnail"]
            ):
                self_.thumbnail = kvargs["thumbnail"]
            else:
                self_.thumbnail = Link(**kvargs["thumbnail"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemMetaResponseBody:
    """
    Item metadata and computed fields.

    Attributes
    ----------
    actions: list[str]
      The actions that the user can perform on the item.
    collections: list[ItemTagResponseBody]
      An array of collections that the item is part of.
    isFavorited: bool
      The flag that indicates if item is in the user's favorites collection.
    tags: list[ItemTagResponseBody]
      An array of tags that the item is part of.
    """

    actions: list[str] = None
    collections: list[ItemTagResponseBody] = None
    isFavorited: bool = None
    tags: list[ItemTagResponseBody] = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "collections" in kvargs and kvargs["collections"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemMetaResponseBody.__annotations__["collections"]
                for e in kvargs["collections"]
            ):
                self_.collections = kvargs["collections"]
            else:
                self_.collections = [
                    ItemTagResponseBody(**e) for e in kvargs["collections"]
                ]
        if "isFavorited" in kvargs and kvargs["isFavorited"] is not None:
            self_.isFavorited = kvargs["isFavorited"]
        if "tags" in kvargs and kvargs["tags"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemMetaResponseBody.__annotations__["tags"]
                for e in kvargs["tags"]
            ):
                self_.tags = kvargs["tags"]
            else:
                self_.tags = [ItemTagResponseBody(**e) for e in kvargs["tags"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemResultResponseBody:
    """
    An item.

    Attributes
    ----------
    actions: list[str]
      The actions that the user can perform on the item.
    collectionIds: list[str]
      The ID of the collections that the item has been added to.
    createdAt: str
      The RFC3339 datetime when the item was created.
    creatorId: str
      The ID of the user who created the item. This is only populated if the JWT contains a userId.
    description: str
    id: str
      The item's unique identifier.
    isFavorited: bool
      The flag that indicates if item is in the user's favorites collection.
    itemViews: ItemViewsResponseBody
    links: ItemLinksResponseBody
    meta: ItemMetaResponseBody
      Item metadata and computed fields.
    name: str
    ownerId: str
      The ID of the user who owns the item.
    resourceAttributes: object
    resourceCreatedAt: str
      The RFC3339 datetime when the resource that the item references was created.
    resourceCustomAttributes: object
    resourceId: str
      The case-sensitive string used to search for an item by resourceId. If resourceId is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceLink: str
      The case-sensitive string used to search for an item by resourceLink. If resourceLink is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceReloadEndTime: str
      The RFC3339 datetime when the resource last reload ended.
    resourceReloadStatus: str
      If the resource last reload was successful or not.
    resourceSize: ItemsResourceSizeResponseBody
    resourceSubType: str
      Optional field defining the item's subtype, if any.
    resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"]
      The case-sensitive string defining the item's type.
    resourceUpdatedAt: str
      The RFC3339 datetime when the resource that the item references was last updated.
    spaceId: str
      The space's unique identifier.
    tenantId: str
      The ID of the tenant that owns the item. This is populated using the JWT.
    thumbnailId: str
      The item thumbnail's unique identifier. This is optional for internal resources.
    updatedAt: str
      The RFC3339 datetime when the item was last updated.
    updaterId: str
      ID of the user who last updated the item. This is only populated if the JWT contains a userId.
    """

    actions: list[str] = None
    collectionIds: list[str] = None
    createdAt: str = None
    creatorId: str = None
    description: str = None
    id: str = None
    isFavorited: bool = None
    itemViews: ItemViewsResponseBody = None
    links: ItemLinksResponseBody = None
    meta: ItemMetaResponseBody = None
    name: str = None
    ownerId: str = None
    resourceAttributes: object = None
    resourceCreatedAt: str = None
    resourceCustomAttributes: object = None
    resourceId: str = None
    resourceLink: str = None
    resourceReloadEndTime: str = None
    resourceReloadStatus: str = None
    resourceSize: ItemsResourceSizeResponseBody = None
    resourceSubType: str = None
    resourceType: Literal[
        "app",
        "collection",
        "qlikview",
        "insight",
        "qvapp",
        "genericlink",
        "sharingservicetask",
        "note",
        "dataasset",
        "dataset",
        "automation",
        "automl-experiment",
        "automl-deployment",
    ] = None
    resourceUpdatedAt: str = None
    spaceId: str = None
    tenantId: str = None
    thumbnailId: str = None
    updatedAt: str = None
    updaterId: str = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "collectionIds" in kvargs and kvargs["collectionIds"] is not None:
            self_.collectionIds = kvargs["collectionIds"]
        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "creatorId" in kvargs and kvargs["creatorId"] is not None:
            self_.creatorId = kvargs["creatorId"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "isFavorited" in kvargs and kvargs["isFavorited"] is not None:
            self_.isFavorited = kvargs["isFavorited"]
        if "itemViews" in kvargs and kvargs["itemViews"] is not None:
            if (
                type(kvargs["itemViews"]).__name__
                == ItemResultResponseBody.__annotations__["itemViews"]
            ):
                self_.itemViews = kvargs["itemViews"]
            else:
                self_.itemViews = ItemViewsResponseBody(**kvargs["itemViews"])
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ItemLinksResponseBody(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if (
                type(kvargs["meta"]).__name__
                == ItemResultResponseBody.__annotations__["meta"]
            ):
                self_.meta = kvargs["meta"]
            else:
                self_.meta = ItemMetaResponseBody(**kvargs["meta"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "ownerId" in kvargs and kvargs["ownerId"] is not None:
            self_.ownerId = kvargs["ownerId"]
        if "resourceAttributes" in kvargs and kvargs["resourceAttributes"] is not None:
            self_.resourceAttributes = kvargs["resourceAttributes"]
        if "resourceCreatedAt" in kvargs and kvargs["resourceCreatedAt"] is not None:
            self_.resourceCreatedAt = kvargs["resourceCreatedAt"]
        if (
            "resourceCustomAttributes" in kvargs
            and kvargs["resourceCustomAttributes"] is not None
        ):
            self_.resourceCustomAttributes = kvargs["resourceCustomAttributes"]
        if "resourceId" in kvargs and kvargs["resourceId"] is not None:
            self_.resourceId = kvargs["resourceId"]
        if "resourceLink" in kvargs and kvargs["resourceLink"] is not None:
            self_.resourceLink = kvargs["resourceLink"]
        if (
            "resourceReloadEndTime" in kvargs
            and kvargs["resourceReloadEndTime"] is not None
        ):
            self_.resourceReloadEndTime = kvargs["resourceReloadEndTime"]
        if (
            "resourceReloadStatus" in kvargs
            and kvargs["resourceReloadStatus"] is not None
        ):
            self_.resourceReloadStatus = kvargs["resourceReloadStatus"]
        if "resourceSize" in kvargs and kvargs["resourceSize"] is not None:
            if (
                type(kvargs["resourceSize"]).__name__
                == ItemResultResponseBody.__annotations__["resourceSize"]
            ):
                self_.resourceSize = kvargs["resourceSize"]
            else:
                self_.resourceSize = ItemsResourceSizeResponseBody(
                    **kvargs["resourceSize"]
                )
        if "resourceSubType" in kvargs and kvargs["resourceSubType"] is not None:
            self_.resourceSubType = kvargs["resourceSubType"]
        if "resourceType" in kvargs and kvargs["resourceType"] is not None:
            self_.resourceType = kvargs["resourceType"]
        if "resourceUpdatedAt" in kvargs and kvargs["resourceUpdatedAt"] is not None:
            self_.resourceUpdatedAt = kvargs["resourceUpdatedAt"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "thumbnailId" in kvargs and kvargs["thumbnailId"] is not None:
            self_.thumbnailId = kvargs["thumbnailId"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "updaterId" in kvargs and kvargs["updaterId"] is not None:
            self_.updaterId = kvargs["updaterId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemTagResponseBody:
    """
    Holds basic information about a tag or collection.

    Attributes
    ----------
    id: str
      The ID of the tag/collection.
    name: str
      The name of the tag/collection.
    """

    id: str = None
    name: str = None

    def __init__(self_, **kvargs):

        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemViewsResponseBody:
    """

    Attributes
    ----------
    total: int
      Total number of views the resource got during the last 28 days.
    trend: float
      Trend in views over the last 4 weeks.
    unique: int
      Number of unique users who viewed the resource during the last 28 days.
    usedBy: int
      Number of apps this dataset is used in (datasets only).
    week: list[ItemViewsWeeksResponseBody]
    """

    total: int = None
    trend: float = None
    unique: int = None
    usedBy: int = None
    week: list[ItemViewsWeeksResponseBody] = None

    def __init__(self_, **kvargs):

        if "total" in kvargs and kvargs["total"] is not None:
            self_.total = kvargs["total"]
        if "trend" in kvargs and kvargs["trend"] is not None:
            self_.trend = kvargs["trend"]
        if "unique" in kvargs and kvargs["unique"] is not None:
            self_.unique = kvargs["unique"]
        if "usedBy" in kvargs and kvargs["usedBy"] is not None:
            self_.usedBy = kvargs["usedBy"]
        if "week" in kvargs and kvargs["week"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemViewsResponseBody.__annotations__["week"]
                for e in kvargs["week"]
            ):
                self_.week = kvargs["week"]
            else:
                self_.week = [ItemViewsWeeksResponseBody(**e) for e in kvargs["week"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemViewsWeeksResponseBody:
    """

    Attributes
    ----------
    start: str
      The RFC3339 datetime representing the start of the referenced week.
    total: int
      Total number of views the resource got during the referenced week.
    unique: int
      Number of unique users who viewed the resource during the referenced week.
    """

    start: str = None
    total: int = None
    unique: int = None

    def __init__(self_, **kvargs):

        if "start" in kvargs and kvargs["start"] is not None:
            self_.start = kvargs["start"]
        if "total" in kvargs and kvargs["total"] is not None:
            self_.total = kvargs["total"]
        if "unique" in kvargs and kvargs["unique"] is not None:
            self_.unique = kvargs["unique"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsLinksResponseBody:
    """

    Attributes
    ----------
    collection: Link
    next: Link
    prev: Link
    self: Link
    """

    collection: Link = None
    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "collection" in kvargs and kvargs["collection"] is not None:
            if (
                type(kvargs["collection"]).__name__
                == ItemsLinksResponseBody.__annotations__["collection"]
            ):
                self_.collection = kvargs["collection"]
            else:
                self_.collection = Link(**kvargs["collection"])
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == ItemsLinksResponseBody.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == ItemsLinksResponseBody.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == ItemsLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsResourceSizeResponseBody:
    """

    Attributes
    ----------
    appFile: float
      Size of the app on disk in bytes.
    appMemory: float
      Size of the app in memory in bytes.
    """

    appFile: float = None
    appMemory: float = None

    def __init__(self_, **kvargs):

        if "appFile" in kvargs and kvargs["appFile"] is not None:
            self_.appFile = kvargs["appFile"]
        if "appMemory" in kvargs and kvargs["appMemory"] is not None:
            self_.appMemory = kvargs["appMemory"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsResultResponseBody:
    """
    Multiple items.

    Attributes
    ----------
    data: list[ItemResultResponseBody]
    links: ItemsLinksResponseBody
    """

    data: list[ItemResultResponseBody] = None
    links: ItemsLinksResponseBody = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemsResultResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [ItemResultResponseBody(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemsResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ItemsLinksResponseBody(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Collections:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_favorite(self) -> any:
        """
        Retrieves the user's favorites collection.
        Finds and returns the user's favorites collection.


        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/collections/favorites",
            method="GET",
            params={},
            data=None,
        )
        return response.json()

    def get(self, collectionId: str) -> CollectionResultResponseBody:
        """
        Returns a collection.
        Finds and returns a collection.


        Parameters
        ----------
        collectionId: str
          The collection's unique identifier.
        """
        response = self.auth.rest(
            path="/collections/{collectionId}".replace("{collectionId}", collectionId),
            method="GET",
            params={},
            data=None,
        )
        obj = CollectionResultResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def get_collections(
        self,
        creatorId: str = None,
        id: str = None,
        includeItems: str = None,
        limit: int = None,
        name: str = None,
        next: str = None,
        prev: str = None,
        query: str = None,
        sort: Literal[
            "+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"
        ] = None,
        type: Literal["public", "private"] = None,
        max_items: int = 10,
    ) -> ListableResource[CollectionResultResponseBody]:
        """
        Retrieves collections that the user has access to.
        Finds and returns the collections that the user can access. This endpoint does not return the user's favorites collection.


        Parameters
        ----------
        creatorId: str = None
          The case-sensitive string used to search for a resource by creatorId.
        id: str = None
          The collection's unique identifier.
        includeItems: str = None
          JSON-formatted string for querying collection items.
        limit: int = None
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        name: str = None
          The case-sensitive string used to search for a collection by name.
        next: str = None
          The cursor to the next page of resources. Provide either the
          next or prev cursor, but not both.

        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        query: str = None
          The case-insensitive string used to search for a resource by name or description.
        sort: Literal["+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"] = None
          The property of a resource to sort on (default sort is +createdAt).
          The supported properties are createdAt, updatedAt, and name. A property
          must be prefixed by + or - to indicate ascending or descending sort order
          respectively.

        type: Literal["public", "private"] = None
          The case-sensitive string used to search for a collection by type.
        """
        query_params = {}
        if creatorId is not None:
            query_params["creatorId"] = creatorId
        if id is not None:
            query_params["id"] = id
        if includeItems is not None:
            query_params["includeItems"] = includeItems
        if limit is not None:
            query_params["limit"] = limit
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if query is not None:
            query_params["query"] = query
        if sort is not None:
            query_params["sort"] = sort
        if type is not None:
            query_params["type"] = type
        response = self.auth.rest(
            path="/collections",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=CollectionResultResponseBody,
            auth=self.auth,
            path="/collections",
            max_items=max_items,
            query_params=query_params,
        )

    def create(
        self, data: CollectionsCreateCollectionRequestBody
    ) -> CollectionResultResponseBody:
        """
        Creates a new collection.
        Creates and returns a new collection. Collections can have the same name.


        Parameters
        ----------
        data: CollectionsCreateCollectionRequestBody

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/collections",
            method="POST",
            params={},
            data=data,
        )
        obj = CollectionResultResponseBody(**response.json())
        obj.auth = self.auth
        return obj



import base64
import hashlib
import os
import random
import re
import string
from typing import Dict, List
from urllib.parse import urlencode

from .config import Config


def _generate_code_verifier() -> str:
    """
    method for generating random alphanumeric string, only to be used in the PKCE flow
    """
    code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
    code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
    return code_verifier


def _generate_code_challenge() -> List[str]:
    """
    Create a code challenge
    returns [code_challenge, code_verifier]
    """
    code_verifier = _generate_code_verifier()
    code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
    code_challenge = code_challenge.replace("=", "")
    return code_challenge, code_verifier


def _generate_authorization_url(
    config: Config, state: str = None, code_challenge: bool = False
) -> Dict[str, str]:
    if state is None:
        state = _generate_random_string(7)

    authorize_path = f"{config.host.strip('/')}/oauth/authorize"
    query_params = {
        "response_type": "code",
        "client_id": config.client_id,
        "redirect_uri": config.redirect_url,
        "scope": " ".join(config.scope),
        "state": state,
    }
    code_verifier = None
    if code_challenge is True:
        code_challenge, code_verifier = _generate_code_challenge()
        query_params["code_challenge"] = code_challenge
        query_params["code_challenge_method"] = "S256"

    url = authorize_path + "?" + urlencode(query_params)
    return {"url": url, "code_verifier": code_verifier, "state": state}


def _generate_random_string(string_length: int = 7):
    return "".join(
        random.choices(string.ascii_lowercase + string.digits, k=string_length)
    )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
      The URL for the link
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class OauthToken:
    """

    Attributes
    ----------
    description: str
      The description of the token
    deviceType: str
      The type of the user device the authorization token is generated for (Tablet, Phone etc.)
    id: str
      The token ID
    lastUsed: str
      The last time the token was used
    tenantId: str
      The ID of the owning tenant
    userId: str
      The ID of the owning user
    """

    description: str = None
    deviceType: str = None
    id: str = None
    lastUsed: str = None
    tenantId: str = None
    userId: str = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "deviceType" in kvargs and kvargs["deviceType"] is not None:
            self_.deviceType = kvargs["deviceType"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastUsed" in kvargs and kvargs["lastUsed"] is not None:
            self_.lastUsed = kvargs["lastUsed"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "userId" in kvargs and kvargs["userId"] is not None:
            self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class OauthTokenPage:
    """

    Attributes
    ----------
    data: list[OauthToken]
    links: OauthTokenPageLinks
    """

    data: list[OauthToken] = None
    links: OauthTokenPageLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == OauthTokenPage.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [OauthToken(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == OauthTokenPage.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = OauthTokenPageLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class OauthTokenPageLinks:
    """

    Attributes
    ----------
    next: Link
    prev: Link
    self: Link
    """

    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == OauthTokenPageLinks.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == OauthTokenPageLinks.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == OauthTokenPageLinks.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class OauthTokens:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def delete(self, tokenId: str) -> None:
        """
        Revoke an OAuth token by ID

        Parameters
        ----------
        tokenId: str
          The ID of the token to revoke
        """
        self.auth.rest(
            path="/oauth-tokens/{tokenId}".replace("{tokenId}", tokenId),
            method="DELETE",
            params={},
            data=None,
        )

    def get_oauth_tokens(
        self,
        limit: float = None,
        page: str = None,
        sort: Literal["userId"] = "userId",
        userId: str = None,
        max_items: int = 10,
    ) -> ListableResource[OauthToken]:
        """
        List OAuth tokens

        Parameters
        ----------
        limit: float = None
          The maximum number of tokens to return
        page: str = None
          The target page
        sort: Literal["userId"] = "userId"
          The field to sort by
        userId: str = None
          The ID of the user to limit results to
        """
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if page is not None:
            query_params["page"] = page
        if sort is not None:
            query_params["sort"] = sort
        if userId is not None:
            query_params["userId"] = userId
        response = self.auth.rest(
            path="/oauth-tokens",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=OauthToken,
            auth=self.auth,
            path="/oauth-tokens",
            max_items=max_items,
            query_params=query_params,
        )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Task:
    """

    Attributes
    ----------
    appId: str
      The ID of the app.
    autoReload: bool
      A flag that indicates whether a reload is triggered when data of the app is changed
    autoReloadPartial: bool
      A flag that indicates whether it is a partial reload or not for the auto reload
    endDateTime: str
      The time that the task will stop recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted.
    partial: bool
      The task is partial reload or not
    recurrence: list[str]
      List of RECUR lines for a recurring event, as specified in RFC5545. Note that DTSTART and DTEND lines are not allowed in this field; event start and end times are specified in the start and end fields. This field is omitted for single events or instances of recurring events
    startDateTime: str
      The time that the task execution start recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted. Field startDateTime should not be before the Unix epoch 00:00:00 UTC on 1 January 1970. Note that the empty string value with the empty recurrence array indicates the scheduled job is not set.
    timeZone: str
      The time zone in which the time is specified. (Formatted as an IANA Time Zone Database name, e.g. Europe/Zurich.) This field specifies the time zone in which the event start/end are expanded. If missing the start/end fields must specify a UTC offset in RFC3339 format.
    fortressId: str
      The fortress ID of the application
    id: str
      The ID of the task.
    lastExecutionTime: str
      The last time the task executed.
    links: SelfLink
    log: str
      The reason why the task was disabled.
    nextExecutionTime: str
      The next time the task will execute.
    spaceId: str
      The space ID of the application
    state: Literal["Enabled", "Disabled", "Completed"]
      Toggle for enabling and disabling the reload task
    tenantId: str
      The ID of the tenant who owns the task.
    userId: str
      The ID of the user who owns the task.
    """

    appId: str = None
    autoReload: bool = None
    autoReloadPartial: bool = None
    endDateTime: str = None
    partial: bool = None
    recurrence: list[str] = None
    startDateTime: str = None
    timeZone: str = None
    fortressId: str = None
    id: str = None
    lastExecutionTime: str = None
    links: SelfLink = None
    log: str = None
    nextExecutionTime: str = None
    spaceId: str = None
    state: Literal["Enabled", "Disabled", "Completed"] = None
    tenantId: str = None
    userId: str = None

    def __init__(self_, **kvargs):

        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "autoReload" in kvargs and kvargs["autoReload"] is not None:
            self_.autoReload = kvargs["autoReload"]
        if "autoReloadPartial" in kvargs and kvargs["autoReloadPartial"] is not None:
            self_.autoReloadPartial = kvargs["autoReloadPartial"]
        if "endDateTime" in kvargs and kvargs["endDateTime"] is not None:
            self_.endDateTime = kvargs["endDateTime"]
        if "partial" in kvargs and kvargs["partial"] is not None:
            self_.partial = kvargs["partial"]
        if "recurrence" in kvargs and kvargs["recurrence"] is not None:
            self_.recurrence = kvargs["recurrence"]
        if "startDateTime" in kvargs and kvargs["startDateTime"] is not None:
            self_.startDateTime = kvargs["startDateTime"]
        if "timeZone" in kvargs and kvargs["timeZone"] is not None:
            self_.timeZone = kvargs["timeZone"]
        if "fortressId" in kvargs and kvargs["fortressId"] is not None:
            self_.fortressId = kvargs["fortressId"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "lastExecutionTime" in kvargs and kvargs["lastExecutionTime"] is not None:
            self_.lastExecutionTime = kvargs["lastExecutionTime"]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Task.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = SelfLink(**kvargs["links"])
        if "log" in kvargs and kvargs["log"] is not None:
            self_.log = kvargs["log"]
        if "nextExecutionTime" in kvargs and kvargs["nextExecutionTime"] is not None:
            self_.nextExecutionTime = kvargs["nextExecutionTime"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "state" in kvargs and kvargs["state"] is not None:
            self_.state = kvargs["state"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "userId" in kvargs and kvargs["userId"] is not None:
            self_.userId = kvargs["userId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Delete a task.


        Parameters
        ----------
        """
        self.auth.rest(
            path="/reload-tasks/{taskId}".replace("{taskId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def set(self, data: PutTaskBody) -> Task:
        """
        Update an existing task.


        Parameters
        ----------
        data: PutTaskBody
          Request body specifying the task parameters.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/reload-tasks/{taskId}".replace("{taskId}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self


@dataclass
class Href:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class PaginationLinks:
    """

    Attributes
    ----------
    self: Href
    next: Href
    prev: Href
    """

    self: Href = None
    next: Href = None
    prev: Href = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == PaginationLinks.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == PaginationLinks.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = Href(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == PaginationLinks.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Href(**kvargs["prev"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class PostTaskBody:
    """

    Attributes
    ----------
    appId: str
      The ID of the app.
    autoReload: bool
      A flag that indicates whether a reload is triggered when data of the app is changed
    autoReloadPartial: bool
      A flag that indicates whether it is a partial reload or not for the auto reload
    endDateTime: str
      The time that the task will stop recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted.
    partial: bool
      The task is partial reload or not
    recurrence: list[str]
      List of RECUR lines for a recurring event, as specified in RFC5545. Note that DTSTART and DTEND lines are not allowed in this field; event start and end times are specified in the start and end fields. This field is omitted for single events or instances of recurring events
    startDateTime: str
      The time that the task execution start recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted. Field startDateTime should not be before the Unix epoch 00:00:00 UTC on 1 January 1970. Note that the empty string value with the empty recurrence array indicates the scheduled job is not set.
    timeZone: str
      The time zone in which the time is specified. (Formatted as an IANA Time Zone Database name, e.g. Europe/Zurich.) This field specifies the time zone in which the event start/end are expanded. If missing the start/end fields must specify a UTC offset in RFC3339 format.
    type: Literal["scheduled_reload"]
      Type of task being created - only contains the "scheduled_reload" value. Type value is not used for creating a schedule reload. It has been deprecated since 2022-04-05.
    """

    appId: str = None
    autoReload: bool = None
    autoReloadPartial: bool = None
    endDateTime: str = None
    partial: bool = None
    recurrence: list[str] = None
    startDateTime: str = None
    timeZone: str = None
    type: Literal["scheduled_reload"] = None

    def __init__(self_, **kvargs):

        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "autoReload" in kvargs and kvargs["autoReload"] is not None:
            self_.autoReload = kvargs["autoReload"]
        if "autoReloadPartial" in kvargs and kvargs["autoReloadPartial"] is not None:
            self_.autoReloadPartial = kvargs["autoReloadPartial"]
        if "endDateTime" in kvargs and kvargs["endDateTime"] is not None:
            self_.endDateTime = kvargs["endDateTime"]
        if "partial" in kvargs and kvargs["partial"] is not None:
            self_.partial = kvargs["partial"]
        if "recurrence" in kvargs and kvargs["recurrence"] is not None:
            self_.recurrence = kvargs["recurrence"]
        if "startDateTime" in kvargs and kvargs["startDateTime"] is not None:
            self_.startDateTime = kvargs["startDateTime"]
        if "timeZone" in kvargs and kvargs["timeZone"] is not None:
            self_.timeZone = kvargs["timeZone"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class PutTaskBody:
    """

    Attributes
    ----------
    appId: str
      The ID of the app.
    autoReload: bool
      A flag that indicates whether a reload is triggered when data of the app is changed
    autoReloadPartial: bool
      A flag that indicates whether it is a partial reload or not for the auto reload
    endDateTime: str
      The time that the task will stop recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted.
    partial: bool
      The task is partial reload or not
    recurrence: list[str]
      List of RECUR lines for a recurring event, as specified in RFC5545. Note that DTSTART and DTEND lines are not allowed in this field; event start and end times are specified in the start and end fields. This field is omitted for single events or instances of recurring events
    startDateTime: str
      The time that the task execution start recurring. If the time zone is missing, this is a combined date-time value expressing a time with a fixed UTC offset (formatted according to RFC3339). If a time zone is given, the zone offset must be omitted. Field startDateTime should not be before the Unix epoch 00:00:00 UTC on 1 January 1970. Note that the empty string value with the empty recurrence array indicates the scheduled job is not set.
    timeZone: str
      The time zone in which the time is specified. (Formatted as an IANA Time Zone Database name, e.g. Europe/Zurich.) This field specifies the time zone in which the event start/end are expanded. If missing the start/end fields must specify a UTC offset in RFC3339 format.
    state: Literal["Enabled", "Disabled", "Completed"]
      Toggle for enabling and disabling the reload task
    """

    appId: str = None
    autoReload: bool = None
    autoReloadPartial: bool = None
    endDateTime: str = None
    partial: bool = None
    recurrence: list[str] = None
    startDateTime: str = None
    timeZone: str = None
    state: Literal["Enabled", "Disabled", "Completed"] = None

    def __init__(self_, **kvargs):

        if "appId" in kvargs and kvargs["appId"] is not None:
            self_.appId = kvargs["appId"]
        if "autoReload" in kvargs and kvargs["autoReload"] is not None:
            self_.autoReload = kvargs["autoReload"]
        if "autoReloadPartial" in kvargs and kvargs["autoReloadPartial"] is not None:
            self_.autoReloadPartial = kvargs["autoReloadPartial"]
        if "endDateTime" in kvargs and kvargs["endDateTime"] is not None:
            self_.endDateTime = kvargs["endDateTime"]
        if "partial" in kvargs and kvargs["partial"] is not None:
            self_.partial = kvargs["partial"]
        if "recurrence" in kvargs and kvargs["recurrence"] is not None:
            self_.recurrence = kvargs["recurrence"]
        if "startDateTime" in kvargs and kvargs["startDateTime"] is not None:
            self_.startDateTime = kvargs["startDateTime"]
        if "timeZone" in kvargs and kvargs["timeZone"] is not None:
            self_.timeZone = kvargs["timeZone"]
        if "state" in kvargs and kvargs["state"] is not None:
            self_.state = kvargs["state"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class SelfLink:
    """

    Attributes
    ----------
    self: Href
    """

    self: Href = None

    def __init__(self_, **kvargs):

        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == SelfLink.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = Href(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Tasks:
    """

    Attributes
    ----------
    data: list[Task]
    links: PaginationLinks
    """

    data: list[Task] = None
    links: PaginationLinks = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == Tasks.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [Task(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == Tasks.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = PaginationLinks(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class ReloadTasks:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get(self, taskId: str) -> Task:
        """
        Find and return a task.


        Parameters
        ----------
        taskId: str
          The unique identifier of the task.
        """
        response = self.auth.rest(
            path="/reload-tasks/{taskId}".replace("{taskId}", taskId),
            method="GET",
            params={},
            data=None,
        )
        obj = Task(**response.json())
        obj.auth = self.auth
        return obj

    def get_reload_tasks(
        self,
        appId: str = None,
        limit: int = 10,
        next: str = None,
        partial: bool = None,
        prev: str = None,
        max_items: int = 10,
    ) -> ListableResource[Task]:
        """
        Find and return the tasks that the user can access.


        Parameters
        ----------
        appId: str = None
          The case sensitive string used to search for a task by app ID.
        limit: int = 10
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        partial: bool = None
          The boolean value used to search for a task is partial or not
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        """
        query_params = {}
        if appId is not None:
            query_params["appId"] = appId
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if partial is not None:
            query_params["partial"] = partial
        if prev is not None:
            query_params["prev"] = prev
        response = self.auth.rest(
            path="/reload-tasks",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Task,
            auth=self.auth,
            path="/reload-tasks",
            max_items=max_items,
            query_params=query_params,
        )

    def create(self, data: PostTaskBody) -> Task:
        """
        Create a task for a specified app.


        Parameters
        ----------
        data: PostTaskBody
          Request body specifying the task parameters.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/reload-tasks",
            method="POST",
            params={},
            data=data,
        )
        obj = Task(**response.json())
        obj.auth = self.auth
        return obj



# This is spectacularly generated code by spectacular v0.0.0 based on
# Qlik Cloud Services Grafana 1.0.0-202205031030

from __future__ import annotations

import io
import json
from dataclasses import asdict, dataclass

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class Theme:
    """
    The theme model.

    Attributes
    ----------
    author: str
      Author of the theme.
    createdAt: str
    dependencies: object
      Map of dependencies describing version of the component it requires.
    description: str
      Description of the theme.
    file: object
      The file that was uploaded with the theme.
    homepage: str
      Home page of the theme.
    icon: str
      Icon to show in the client.
    id: str
    keywords: str
      Keywords for the theme.
    license: str
      Under which license this theme is published.
    name: str
      The display name of this theme.
    qextFilename: str
      The name of the qext file that was uploaded with this theme.
    qextVersion: str
      The version from the qext file that was uploaded with this extension.
    repository: str
      Link to the theme source code.
    supplier: str
      Supplier of the theme.
    tags: list[str]
      List of tags.
    tenantId: str
    type: str
      The type of this theme (visualization, etc.).
    updateAt: str
    userId: str
    version: str
      Version of the theme.
    """

    author: str = None
    createdAt: str = None
    dependencies: object = None
    description: str = None
    file: object = None
    homepage: str = None
    icon: str = None
    id: str = None
    keywords: str = None
    license: str = None
    name: str = None
    qextFilename: str = None
    qextVersion: str = None
    repository: str = None
    supplier: str = None
    tags: list[str] = None
    tenantId: str = None
    type: str = None
    updateAt: str = None
    userId: str = None
    version: str = None

    def __init__(self_, **kvargs):
        if "author" in kvargs:
            if type(kvargs["author"]).__name__ is self_.__annotations__["author"]:
                self_.author = kvargs["author"]
            else:
                self_.author = kvargs["author"]
        if "createdAt" in kvargs:
            if type(kvargs["createdAt"]).__name__ is self_.__annotations__["createdAt"]:
                self_.createdAt = kvargs["createdAt"]
            else:
                self_.createdAt = kvargs["createdAt"]
        if "dependencies" in kvargs:
            if (
                type(kvargs["dependencies"]).__name__
                is self_.__annotations__["dependencies"]
            ):
                self_.dependencies = kvargs["dependencies"]
            else:
                self_.dependencies = kvargs["dependencies"]
        if "description" in kvargs:
            if (
                type(kvargs["description"]).__name__
                is self_.__annotations__["description"]
            ):
                self_.description = kvargs["description"]
            else:
                self_.description = kvargs["description"]
        if "file" in kvargs:
            if type(kvargs["file"]).__name__ is self_.__annotations__["file"]:
                self_.file = kvargs["file"]
            else:
                self_.file = kvargs["file"]
        if "homepage" in kvargs:
            if type(kvargs["homepage"]).__name__ is self_.__annotations__["homepage"]:
                self_.homepage = kvargs["homepage"]
            else:
                self_.homepage = kvargs["homepage"]
        if "icon" in kvargs:
            if type(kvargs["icon"]).__name__ is self_.__annotations__["icon"]:
                self_.icon = kvargs["icon"]
            else:
                self_.icon = kvargs["icon"]
        if "id" in kvargs:
            if type(kvargs["id"]).__name__ is self_.__annotations__["id"]:
                self_.id = kvargs["id"]
            else:
                self_.id = kvargs["id"]
        if "keywords" in kvargs:
            if type(kvargs["keywords"]).__name__ is self_.__annotations__["keywords"]:
                self_.keywords = kvargs["keywords"]
            else:
                self_.keywords = kvargs["keywords"]
        if "license" in kvargs:
            if type(kvargs["license"]).__name__ is self_.__annotations__["license"]:
                self_.license = kvargs["license"]
            else:
                self_.license = kvargs["license"]
        if "name" in kvargs:
            if type(kvargs["name"]).__name__ is self_.__annotations__["name"]:
                self_.name = kvargs["name"]
            else:
                self_.name = kvargs["name"]
        if "qextFilename" in kvargs:
            if (
                type(kvargs["qextFilename"]).__name__
                is self_.__annotations__["qextFilename"]
            ):
                self_.qextFilename = kvargs["qextFilename"]
            else:
                self_.qextFilename = kvargs["qextFilename"]
        if "qextVersion" in kvargs:
            if (
                type(kvargs["qextVersion"]).__name__
                is self_.__annotations__["qextVersion"]
            ):
                self_.qextVersion = kvargs["qextVersion"]
            else:
                self_.qextVersion = kvargs["qextVersion"]
        if "repository" in kvargs:
            if (
                type(kvargs["repository"]).__name__
                is self_.__annotations__["repository"]
            ):
                self_.repository = kvargs["repository"]
            else:
                self_.repository = kvargs["repository"]
        if "supplier" in kvargs:
            if type(kvargs["supplier"]).__name__ is self_.__annotations__["supplier"]:
                self_.supplier = kvargs["supplier"]
            else:
                self_.supplier = kvargs["supplier"]
        if "tags" in kvargs:
            if type(kvargs["tags"]).__name__ is self_.__annotations__["tags"]:
                self_.tags = kvargs["tags"]
            else:
                self_.tags = kvargs["tags"]
        if "tenantId" in kvargs:
            if type(kvargs["tenantId"]).__name__ is self_.__annotations__["tenantId"]:
                self_.tenantId = kvargs["tenantId"]
            else:
                self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs:
            if type(kvargs["type"]).__name__ is self_.__annotations__["type"]:
                self_.type = kvargs["type"]
            else:
                self_.type = kvargs["type"]
        if "updateAt" in kvargs:
            if type(kvargs["updateAt"]).__name__ is self_.__annotations__["updateAt"]:
                self_.updateAt = kvargs["updateAt"]
            else:
                self_.updateAt = kvargs["updateAt"]
        if "userId" in kvargs:
            if type(kvargs["userId"]).__name__ is self_.__annotations__["userId"]:
                self_.userId = kvargs["userId"]
            else:
                self_.userId = kvargs["userId"]
        if "version" in kvargs:
            if type(kvargs["version"]).__name__ is self_.__annotations__["version"]:
                self_.version = kvargs["version"]
            else:
                self_.version = kvargs["version"]
        for k, v in kvargs.items():
            if k not in self_.__annotations__:
                self_.__setattr__(k, v)

    def delete(self) -> None:
        """
        Deletes a specific theme.

        Parameters
        ----------
        """

        self.auth.rest(
            path="/themes/{id}".replace("{id}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def patchs(
        self, file: io.BufferedReader, data: Theme = None, max_items: int = 10
    ) -> ListableResource[Theme]:
        """
        Updates a specific theme with provided data. If a file is provided, the data field is not required.

        Parameters
        ----------
        """

        files_dict = {}
        if data is not None:
            try:
                data = asdict(data)
            except:
                pass
            files_dict["data"] = (None, json.dumps(data))

        files_dict["file"] = file

        response = self.auth.rest(
            path="/themes/{id}".replace("{id}", self.id),
            method="PATCH",
            params={},
            data=None,
            files=files_dict,
        )
        return ListableResource(
            response=response.json(),
            cls=Theme,
            auth=self.auth,
            path="/themes/{id}".replace("{id}", self.id),
            max_items=max_items,
            query_params={},
        )

    def get_file(self) -> None:
        """
        Downloads the theme as an archive.

        Parameters
        ----------
        """

        self.auth.rest(
            path="/themes/{id}/file".replace("{id}", self.id),
            method="GET",
            params={},
            data=None,
        )


@dataclass
class ThemesClass:
    """

    Attributes
    ----------
    data: list[Theme]
    """

    data: list[Theme] = None

    def __init__(self_, **kvargs):
        if "data" in kvargs:
            if type(kvargs["data"]).__name__ is self_.__annotations__["data"]:
                self_.data = kvargs["data"]
            else:
                self_.data = [Theme(**e) for e in kvargs["data"]]
        for k, v in kvargs.items():
            if k not in self_.__annotations__:
                self_.__setattr__(k, v)


class Themes:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_themes(self, max_items: int = 10) -> ListableResource[Theme]:
        """
        Lists all themes.


        Parameters
        ----------
        """

        response = self.auth.rest(
            path="/themes",
            method="GET",
            params={},
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=Theme,
            auth=self.auth,
            path="/themes",
            max_items=max_items,
            query_params={},
        )

    def create(self, file: io.BufferedReader, data: Theme = None) -> Theme:
        """
        Creates a new theme. If a file is provided, the data field is not required.


        Parameters
        ----------
        """

        files_dict = {}
        if data is not None:
            try:
                data = asdict(data)
            except:
                pass
            files_dict["data"] = (None, json.dumps(data))

        files_dict["file"] = file

        response = self.auth.rest(
            path="/themes", method="POST", params={}, data=None, files=files_dict
        )
        obj = Theme(**response.json())
        obj.auth = self.auth
        return obj

    def get(self, id: str) -> Theme:
        """
        Returns a specific theme.


        id: str
          Theme identifier or its qextFilename

        Parameters
        ----------
        id: str
        """

        response = self.auth.rest(
            path="/themes/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = Theme(**response.json())
        obj.auth = self.auth
        return obj

    def get_file(self, id: str, filepath: str) -> None:
        """
        Downloads a file from the theme archive.


        id: str
          Theme identifier or its qextFilename.

        filepath: str
          Path to the file archive for the specified theme archive. Folders separated with forward slashes.

        Parameters
        ----------
        id: str
        filepath: str
        """

        self.auth.rest(
            path="/themes/{id}/file/{filepath}".replace("{id}", id).replace(
                "{filepath}", filepath
            ),
            method="GET",
            params={},
            data=None,
        )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class ItemResultResponseBody:
    """
    An item.

    Attributes
    ----------
    actions: list[str]
      The actions that the user can perform on the item.
    collectionIds: list[str]
      The ID of the collections that the item has been added to.
    createdAt: str
      The RFC3339 datetime when the item was created.
    creatorId: str
      The ID of the user who created the item. This is only populated if the JWT contains a userId.
    description: str
    id: str
      The item's unique identifier.
    isFavorited: bool
      The flag that indicates if item is in the user's favorites collection.
    itemViews: ItemViewsResponseBody
    links: ItemLinksResponseBody
    meta: ItemMetaResponseBody
      Item metadata and computed fields.
    name: str
    ownerId: str
      The ID of the user who owns the item.
    resourceAttributes: object
    resourceCreatedAt: str
      The RFC3339 datetime when the resource that the item references was created.
    resourceCustomAttributes: object
    resourceId: str
      The case-sensitive string used to search for an item by resourceId. If resourceId is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceLink: str
      The case-sensitive string used to search for an item by resourceLink. If resourceLink is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceReloadEndTime: str
      The RFC3339 datetime when the resource last reload ended.
    resourceReloadStatus: str
      If the resource last reload was successful or not.
    resourceSize: ItemsResourceSizeResponseBody
    resourceSubType: str
      Optional field defining the item's subtype, if any.
    resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"]
      The case-sensitive string defining the item's type.
    resourceUpdatedAt: str
      The RFC3339 datetime when the resource that the item references was last updated.
    spaceId: str
      The space's unique identifier.
    tenantId: str
      The ID of the tenant that owns the item. This is populated using the JWT.
    thumbnailId: str
      The item thumbnail's unique identifier. This is optional for internal resources.
    updatedAt: str
      The RFC3339 datetime when the item was last updated.
    updaterId: str
      ID of the user who last updated the item. This is only populated if the JWT contains a userId.
    """

    actions: list[str] = None
    collectionIds: list[str] = None
    createdAt: str = None
    creatorId: str = None
    description: str = None
    id: str = None
    isFavorited: bool = None
    itemViews: ItemViewsResponseBody = None
    links: ItemLinksResponseBody = None
    meta: ItemMetaResponseBody = None
    name: str = None
    ownerId: str = None
    resourceAttributes: object = None
    resourceCreatedAt: str = None
    resourceCustomAttributes: object = None
    resourceId: str = None
    resourceLink: str = None
    resourceReloadEndTime: str = None
    resourceReloadStatus: str = None
    resourceSize: ItemsResourceSizeResponseBody = None
    resourceSubType: str = None
    resourceType: Literal[
        "app",
        "collection",
        "qlikview",
        "insight",
        "qvapp",
        "genericlink",
        "sharingservicetask",
        "note",
        "dataasset",
        "dataset",
        "automation",
        "automl-experiment",
        "automl-deployment",
    ] = None
    resourceUpdatedAt: str = None
    spaceId: str = None
    tenantId: str = None
    thumbnailId: str = None
    updatedAt: str = None
    updaterId: str = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "collectionIds" in kvargs and kvargs["collectionIds"] is not None:
            self_.collectionIds = kvargs["collectionIds"]
        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "creatorId" in kvargs and kvargs["creatorId"] is not None:
            self_.creatorId = kvargs["creatorId"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "isFavorited" in kvargs and kvargs["isFavorited"] is not None:
            self_.isFavorited = kvargs["isFavorited"]
        if "itemViews" in kvargs and kvargs["itemViews"] is not None:
            if (
                type(kvargs["itemViews"]).__name__
                == ItemResultResponseBody.__annotations__["itemViews"]
            ):
                self_.itemViews = kvargs["itemViews"]
            else:
                self_.itemViews = ItemViewsResponseBody(**kvargs["itemViews"])
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ItemLinksResponseBody(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if (
                type(kvargs["meta"]).__name__
                == ItemResultResponseBody.__annotations__["meta"]
            ):
                self_.meta = kvargs["meta"]
            else:
                self_.meta = ItemMetaResponseBody(**kvargs["meta"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "ownerId" in kvargs and kvargs["ownerId"] is not None:
            self_.ownerId = kvargs["ownerId"]
        if "resourceAttributes" in kvargs and kvargs["resourceAttributes"] is not None:
            self_.resourceAttributes = kvargs["resourceAttributes"]
        if "resourceCreatedAt" in kvargs and kvargs["resourceCreatedAt"] is not None:
            self_.resourceCreatedAt = kvargs["resourceCreatedAt"]
        if (
            "resourceCustomAttributes" in kvargs
            and kvargs["resourceCustomAttributes"] is not None
        ):
            self_.resourceCustomAttributes = kvargs["resourceCustomAttributes"]
        if "resourceId" in kvargs and kvargs["resourceId"] is not None:
            self_.resourceId = kvargs["resourceId"]
        if "resourceLink" in kvargs and kvargs["resourceLink"] is not None:
            self_.resourceLink = kvargs["resourceLink"]
        if (
            "resourceReloadEndTime" in kvargs
            and kvargs["resourceReloadEndTime"] is not None
        ):
            self_.resourceReloadEndTime = kvargs["resourceReloadEndTime"]
        if (
            "resourceReloadStatus" in kvargs
            and kvargs["resourceReloadStatus"] is not None
        ):
            self_.resourceReloadStatus = kvargs["resourceReloadStatus"]
        if "resourceSize" in kvargs and kvargs["resourceSize"] is not None:
            if (
                type(kvargs["resourceSize"]).__name__
                == ItemResultResponseBody.__annotations__["resourceSize"]
            ):
                self_.resourceSize = kvargs["resourceSize"]
            else:
                self_.resourceSize = ItemsResourceSizeResponseBody(
                    **kvargs["resourceSize"]
                )
        if "resourceSubType" in kvargs and kvargs["resourceSubType"] is not None:
            self_.resourceSubType = kvargs["resourceSubType"]
        if "resourceType" in kvargs and kvargs["resourceType"] is not None:
            self_.resourceType = kvargs["resourceType"]
        if "resourceUpdatedAt" in kvargs and kvargs["resourceUpdatedAt"] is not None:
            self_.resourceUpdatedAt = kvargs["resourceUpdatedAt"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "thumbnailId" in kvargs and kvargs["thumbnailId"] is not None:
            self_.thumbnailId = kvargs["thumbnailId"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "updaterId" in kvargs and kvargs["updaterId"] is not None:
            self_.updaterId = kvargs["updaterId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)

    def get_collections(
        self,
        limit: int = None,
        name: str = None,
        next: str = None,
        prev: str = None,
        query: str = None,
        sort: Literal[
            "+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"
        ] = None,
        type: Literal["public", "private"] = None,
        max_items: int = 10,
    ) -> ListableResource[CollectionResultResponseBody]:
        """
        Returns the collections of an item.
        Finds and returns the collections of an item. This endpoint does not return the user's favorites collection.


        Parameters
        ----------
        limit: int = None
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        name: str = None
          The case-sensitive string used to search for a collection by name.
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        query: str = None
          The case-insensitive string used to search for a resource by name or description.
        sort: Literal["+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"] = None
          The property of a resource to sort on (default sort is +createdAt). The supported properties are createdAt, updatedAt, and name. A property must be prefixed by + or   - to indicate ascending or descending sort order respectively.
        type: Literal["public", "private"] = None
          The case-sensitive string used to search for a collection by type.
        """
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if query is not None:
            query_params["query"] = query
        if sort is not None:
            query_params["sort"] = sort
        if type is not None:
            query_params["type"] = type
        response = self.auth.rest(
            path="/items/{itemId}/collections".replace("{itemId}", self.id),
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=CollectionResultResponseBody,
            auth=self.auth,
            path="/items/{itemId}/collections".replace("{itemId}", self.id),
            max_items=max_items,
            query_params=query_params,
        )

    def get_publisheditems(
        self,
        limit: int = None,
        next: str = None,
        prev: str = None,
        resourceType: Literal[
            "app",
            "collection",
            "qlikview",
            "insight",
            "qvapp",
            "genericlink",
            "sharingservicetask",
            "note",
            "dataasset",
            "dataset",
            "automation",
            "automl-experiment",
            "automl-deployment",
        ] = None,
        sort: Literal[
            "+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"
        ] = None,
        max_items: int = 10,
    ) -> ListableResource[CollectionResultResponseBody]:
        """
        Returns published items for a given item.
        Finds and returns the published items for a given item.


        Parameters
        ----------
        limit: int = None
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"] = None
          The case-sensitive string used to search for an item by resourceType.
        sort: Literal["+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"] = None
          The property of a resource to sort on (default sort is +createdAt). The supported properties are createdAt, updatedAt, and name. A property must be prefixed by + or   - to indicate ascending or descending sort order respectively.
        """
        query_params = {}
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        if resourceType is not None:
            query_params["resourceType"] = resourceType
        if sort is not None:
            query_params["sort"] = sort
        response = self.auth.rest(
            path="/items/{itemId}/publisheditems".replace("{itemId}", self.id),
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=CollectionResultResponseBody,
            auth=self.auth,
            path="/items/{itemId}/publisheditems".replace("{itemId}", self.id),
            max_items=max_items,
            query_params=query_params,
        )

    def delete(self) -> None:
        """
        Deletes an item.
        Deletes an item and removes the item from all collections.


        Parameters
        ----------
        """
        self.auth.rest(
            path="/items/{itemId}".replace("{itemId}", self.id),
            method="DELETE",
            params={},
            data=None,
        )

    def set(self, data: ItemsUpdateItemRequestBody) -> ItemResultResponseBody:
        """
        Updates an item.
        Updates an item. Omitted and unsupported fields are ignored. To unset a field, provide the field's zero value.


        Parameters
        ----------
        data: ItemsUpdateItemRequestBody

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/items/{itemId}".replace("{itemId}", self.id),
            method="PUT",
            params={},
            data=data,
        )
        self.__init__(**response.json())
        return self


@dataclass
class ItemsListItemCollectionsResponseBody:
    """
    ListItemCollectionsResponseBody result type

    Attributes
    ----------
    data: list[CollectionResultResponseBody]
    links: CollectionsLinksResponseBody
    """

    data: list[CollectionResultResponseBody] = None
    links: CollectionsLinksResponseBody = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemsListItemCollectionsResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [CollectionResultResponseBody(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemsListItemCollectionsResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = CollectionsLinksResponseBody(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsListItemsResponseBody:
    """
    ListItemsResponseBody result type

    Attributes
    ----------
    data: list[ItemResultResponseBody]
    links: ItemsLinksResponseBody
    """

    data: list[ItemResultResponseBody] = None
    links: ItemsLinksResponseBody = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemsListItemsResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [ItemResultResponseBody(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemsListItemsResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ItemsLinksResponseBody(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsSettingsPatch:
    """

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsSettingsResponseBody:
    """

    Attributes
    ----------
    usageMetricsEnabled: bool
      Decides if the usage metrics will be shown in the hub UI.
    """

    usageMetricsEnabled: bool = True

    def __init__(self_, **kvargs):

        if (
            "usageMetricsEnabled" in kvargs
            and kvargs["usageMetricsEnabled"] is not None
        ):
            self_.usageMetricsEnabled = kvargs["usageMetricsEnabled"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsUpdateItemRequestBody:
    """

    Attributes
    ----------
    description: str
    name: str
    resourceAttributes: object
    resourceCustomAttributes: object
    resourceId: str
      The case-sensitive string used to search for an item by resourceId. If resourceId is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceLink: str
      The case-sensitive string used to search for an item by resourceLink. If resourceLink is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
    resourceSubType: str
      Optional field defining the item's subtype, if any.
    resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"]
      The case-sensitive string defining the item's type.
    resourceUpdatedAt: str
      The RFC3339 datetime when the resource that the item references was last updated.
    spaceId: str
      The space's unique identifier.
    thumbnailId: str
      The item thumbnail's unique identifier. This is optional for internal resources.
    """

    description: str = None
    name: str = None
    resourceAttributes: object = None
    resourceCustomAttributes: object = None
    resourceId: str = None
    resourceLink: str = None
    resourceSubType: str = None
    resourceType: Literal[
        "app",
        "collection",
        "qlikview",
        "insight",
        "qvapp",
        "genericlink",
        "sharingservicetask",
        "note",
        "dataasset",
        "dataset",
        "automation",
        "automl-experiment",
        "automl-deployment",
    ] = None
    resourceUpdatedAt: str = None
    spaceId: str = None
    thumbnailId: str = None

    def __init__(self_, **kvargs):

        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "resourceAttributes" in kvargs and kvargs["resourceAttributes"] is not None:
            self_.resourceAttributes = kvargs["resourceAttributes"]
        if (
            "resourceCustomAttributes" in kvargs
            and kvargs["resourceCustomAttributes"] is not None
        ):
            self_.resourceCustomAttributes = kvargs["resourceCustomAttributes"]
        if "resourceId" in kvargs and kvargs["resourceId"] is not None:
            self_.resourceId = kvargs["resourceId"]
        if "resourceLink" in kvargs and kvargs["resourceLink"] is not None:
            self_.resourceLink = kvargs["resourceLink"]
        if "resourceSubType" in kvargs and kvargs["resourceSubType"] is not None:
            self_.resourceSubType = kvargs["resourceSubType"]
        if "resourceType" in kvargs and kvargs["resourceType"] is not None:
            self_.resourceType = kvargs["resourceType"]
        if "resourceUpdatedAt" in kvargs and kvargs["resourceUpdatedAt"] is not None:
            self_.resourceUpdatedAt = kvargs["resourceUpdatedAt"]
        if "spaceId" in kvargs and kvargs["spaceId"] is not None:
            self_.spaceId = kvargs["spaceId"]
        if "thumbnailId" in kvargs and kvargs["thumbnailId"] is not None:
            self_.thumbnailId = kvargs["thumbnailId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Link:
    """

    Attributes
    ----------
    href: str
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionLinksResponseBody:
    """

    Attributes
    ----------
    items: Link
    self: Link
    """

    items: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "items" in kvargs and kvargs["items"] is not None:
            if (
                type(kvargs["items"]).__name__
                == CollectionLinksResponseBody.__annotations__["items"]
            ):
                self_.items = kvargs["items"]
            else:
                self_.items = Link(**kvargs["items"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == CollectionLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionMetaResponseBody:
    """
    Collection metadata and computed fields.

    Attributes
    ----------
    items: ItemsResultResponseBody
      Multiple items.
    """

    items: ItemsResultResponseBody = None

    def __init__(self_, **kvargs):

        if "items" in kvargs and kvargs["items"] is not None:
            if (
                type(kvargs["items"]).__name__
                == CollectionMetaResponseBody.__annotations__["items"]
            ):
                self_.items = kvargs["items"]
            else:
                self_.items = ItemsResultResponseBody(**kvargs["items"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionResultResponseBody:
    """
    A collection.

    Attributes
    ----------
    createdAt: str
      The RFC3339 datetime when the collection was created.
    creatorId: str
      The ID of the user who created the collection. This property is only populated if the JWT contains a userId.
    description: str
    id: str
      The collection's unique identifier.
    itemCount: int
      The number of items that have been added to the collection.
    links: CollectionLinksResponseBody
    meta: CollectionMetaResponseBody
      Collection metadata and computed fields.
    name: str
    tenantId: str
      The ID of the tenant that owns the collection. This property is populated by using JWT.
    type: Literal["public", "private", "favorite"]
      The collection's type.
    updatedAt: str
      The RFC3339 datetime when the collection was last updated.
    updaterId: str
      The ID of the user who last updated the collection. This property is only populated if the JWT contains a userId.
    """

    createdAt: str = None
    creatorId: str = None
    description: str = None
    id: str = None
    itemCount: int = None
    links: CollectionLinksResponseBody = None
    meta: CollectionMetaResponseBody = None
    name: str = None
    tenantId: str = None
    type: Literal["public", "private", "favorite"] = None
    updatedAt: str = None
    updaterId: str = None

    def __init__(self_, **kvargs):

        if "createdAt" in kvargs and kvargs["createdAt"] is not None:
            self_.createdAt = kvargs["createdAt"]
        if "creatorId" in kvargs and kvargs["creatorId"] is not None:
            self_.creatorId = kvargs["creatorId"]
        if "description" in kvargs and kvargs["description"] is not None:
            self_.description = kvargs["description"]
        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "itemCount" in kvargs and kvargs["itemCount"] is not None:
            self_.itemCount = kvargs["itemCount"]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == CollectionResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = CollectionLinksResponseBody(**kvargs["links"])
        if "meta" in kvargs and kvargs["meta"] is not None:
            if (
                type(kvargs["meta"]).__name__
                == CollectionResultResponseBody.__annotations__["meta"]
            ):
                self_.meta = kvargs["meta"]
            else:
                self_.meta = CollectionMetaResponseBody(**kvargs["meta"])
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        if "tenantId" in kvargs and kvargs["tenantId"] is not None:
            self_.tenantId = kvargs["tenantId"]
        if "type" in kvargs and kvargs["type"] is not None:
            self_.type = kvargs["type"]
        if "updatedAt" in kvargs and kvargs["updatedAt"] is not None:
            self_.updatedAt = kvargs["updatedAt"]
        if "updaterId" in kvargs and kvargs["updaterId"] is not None:
            self_.updaterId = kvargs["updaterId"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class CollectionsLinksResponseBody:
    """

    Attributes
    ----------
    item: Link
    next: Link
    prev: Link
    self: Link
    """

    item: Link = None
    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "item" in kvargs and kvargs["item"] is not None:
            if (
                type(kvargs["item"]).__name__
                == CollectionsLinksResponseBody.__annotations__["item"]
            ):
                self_.item = kvargs["item"]
            else:
                self_.item = Link(**kvargs["item"])
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == CollectionsLinksResponseBody.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == CollectionsLinksResponseBody.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == CollectionsLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemLinksResponseBody:
    """

    Attributes
    ----------
    collections: Link
    open: Link
    self: Link
    thumbnail: Link
    """

    collections: Link = None
    open: Link = None
    self: Link = None
    thumbnail: Link = None

    def __init__(self_, **kvargs):

        if "collections" in kvargs and kvargs["collections"] is not None:
            if (
                type(kvargs["collections"]).__name__
                == ItemLinksResponseBody.__annotations__["collections"]
            ):
                self_.collections = kvargs["collections"]
            else:
                self_.collections = Link(**kvargs["collections"])
        if "open" in kvargs and kvargs["open"] is not None:
            if (
                type(kvargs["open"]).__name__
                == ItemLinksResponseBody.__annotations__["open"]
            ):
                self_.open = kvargs["open"]
            else:
                self_.open = Link(**kvargs["open"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == ItemLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        if "thumbnail" in kvargs and kvargs["thumbnail"] is not None:
            if (
                type(kvargs["thumbnail"]).__name__
                == ItemLinksResponseBody.__annotations__["thumbnail"]
            ):
                self_.thumbnail = kvargs["thumbnail"]
            else:
                self_.thumbnail = Link(**kvargs["thumbnail"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemMetaResponseBody:
    """
    Item metadata and computed fields.

    Attributes
    ----------
    actions: list[str]
      The actions that the user can perform on the item.
    collections: list[ItemTagResponseBody]
      An array of collections that the item is part of.
    isFavorited: bool
      The flag that indicates if item is in the user's favorites collection.
    tags: list[ItemTagResponseBody]
      An array of tags that the item is part of.
    """

    actions: list[str] = None
    collections: list[ItemTagResponseBody] = None
    isFavorited: bool = None
    tags: list[ItemTagResponseBody] = None

    def __init__(self_, **kvargs):

        if "actions" in kvargs and kvargs["actions"] is not None:
            self_.actions = kvargs["actions"]
        if "collections" in kvargs and kvargs["collections"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemMetaResponseBody.__annotations__["collections"]
                for e in kvargs["collections"]
            ):
                self_.collections = kvargs["collections"]
            else:
                self_.collections = [
                    ItemTagResponseBody(**e) for e in kvargs["collections"]
                ]
        if "isFavorited" in kvargs and kvargs["isFavorited"] is not None:
            self_.isFavorited = kvargs["isFavorited"]
        if "tags" in kvargs and kvargs["tags"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemMetaResponseBody.__annotations__["tags"]
                for e in kvargs["tags"]
            ):
                self_.tags = kvargs["tags"]
            else:
                self_.tags = [ItemTagResponseBody(**e) for e in kvargs["tags"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemTagResponseBody:
    """
    Holds basic information about a tag or collection.

    Attributes
    ----------
    id: str
      The ID of the tag/collection.
    name: str
      The name of the tag/collection.
    """

    id: str = None
    name: str = None

    def __init__(self_, **kvargs):

        if "id" in kvargs and kvargs["id"] is not None:
            self_.id = kvargs["id"]
        if "name" in kvargs and kvargs["name"] is not None:
            self_.name = kvargs["name"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemViewsResponseBody:
    """

    Attributes
    ----------
    total: int
      Total number of views the resource got during the last 28 days.
    trend: float
      Trend in views over the last 4 weeks.
    unique: int
      Number of unique users who viewed the resource during the last 28 days.
    usedBy: int
      Number of apps this dataset is used in (datasets only).
    week: list[ItemViewsWeeksResponseBody]
    """

    total: int = None
    trend: float = None
    unique: int = None
    usedBy: int = None
    week: list[ItemViewsWeeksResponseBody] = None

    def __init__(self_, **kvargs):

        if "total" in kvargs and kvargs["total"] is not None:
            self_.total = kvargs["total"]
        if "trend" in kvargs and kvargs["trend"] is not None:
            self_.trend = kvargs["trend"]
        if "unique" in kvargs and kvargs["unique"] is not None:
            self_.unique = kvargs["unique"]
        if "usedBy" in kvargs and kvargs["usedBy"] is not None:
            self_.usedBy = kvargs["usedBy"]
        if "week" in kvargs and kvargs["week"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemViewsResponseBody.__annotations__["week"]
                for e in kvargs["week"]
            ):
                self_.week = kvargs["week"]
            else:
                self_.week = [ItemViewsWeeksResponseBody(**e) for e in kvargs["week"]]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemViewsWeeksResponseBody:
    """

    Attributes
    ----------
    start: str
      The RFC3339 datetime representing the start of the referenced week.
    total: int
      Total number of views the resource got during the referenced week.
    unique: int
      Number of unique users who viewed the resource during the referenced week.
    """

    start: str = None
    total: int = None
    unique: int = None

    def __init__(self_, **kvargs):

        if "start" in kvargs and kvargs["start"] is not None:
            self_.start = kvargs["start"]
        if "total" in kvargs and kvargs["total"] is not None:
            self_.total = kvargs["total"]
        if "unique" in kvargs and kvargs["unique"] is not None:
            self_.unique = kvargs["unique"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsLinksResponseBody:
    """

    Attributes
    ----------
    collection: Link
    next: Link
    prev: Link
    self: Link
    """

    collection: Link = None
    next: Link = None
    prev: Link = None
    self: Link = None

    def __init__(self_, **kvargs):

        if "collection" in kvargs and kvargs["collection"] is not None:
            if (
                type(kvargs["collection"]).__name__
                == ItemsLinksResponseBody.__annotations__["collection"]
            ):
                self_.collection = kvargs["collection"]
            else:
                self_.collection = Link(**kvargs["collection"])
        if "next" in kvargs and kvargs["next"] is not None:
            if (
                type(kvargs["next"]).__name__
                == ItemsLinksResponseBody.__annotations__["next"]
            ):
                self_.next = kvargs["next"]
            else:
                self_.next = Link(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if (
                type(kvargs["prev"]).__name__
                == ItemsLinksResponseBody.__annotations__["prev"]
            ):
                self_.prev = kvargs["prev"]
            else:
                self_.prev = Link(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if (
                type(kvargs["self"]).__name__
                == ItemsLinksResponseBody.__annotations__["self"]
            ):
                self_.self = kvargs["self"]
            else:
                self_.self = Link(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsResourceSizeResponseBody:
    """

    Attributes
    ----------
    appFile: float
      Size of the app on disk in bytes.
    appMemory: float
      Size of the app in memory in bytes.
    """

    appFile: float = None
    appMemory: float = None

    def __init__(self_, **kvargs):

        if "appFile" in kvargs and kvargs["appFile"] is not None:
            self_.appFile = kvargs["appFile"]
        if "appMemory" in kvargs and kvargs["appMemory"] is not None:
            self_.appMemory = kvargs["appMemory"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class ItemsResultResponseBody:
    """
    Multiple items.

    Attributes
    ----------
    data: list[ItemResultResponseBody]
    links: ItemsLinksResponseBody
    """

    data: list[ItemResultResponseBody] = None
    links: ItemsLinksResponseBody = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == ItemsResultResponseBody.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [ItemResultResponseBody(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if (
                type(kvargs["links"]).__name__
                == ItemsResultResponseBody.__annotations__["links"]
            ):
                self_.links = kvargs["links"]
            else:
                self_.links = ItemsLinksResponseBody(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class Items:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_settings(self) -> ItemsSettingsResponseBody:
        """
        Returns tenant specific settings.
        Finds and returns the settings for the current tenant.


        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/items/settings",
            method="GET",
            params={},
            data=None,
        )
        obj = ItemsSettingsResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def patch_settings(self, data: ItemsSettingsPatch) -> ItemsSettingsResponseBody:
        """
        Patches tenant specific settings.
        Updates the settings provided in the patch body.

        Parameters
        ----------
        data: ItemsSettingsPatch

        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/items/settings",
            method="PATCH",
            params={},
            data=data,
        )
        obj = ItemsSettingsResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def get(self, itemId: str) -> ItemResultResponseBody:
        """
        Returns an item.
        Finds and returns an item.


        Parameters
        ----------
        itemId: str
          The item's unique identifier
        """
        response = self.auth.rest(
            path="/items/{itemId}".replace("{itemId}", itemId),
            method="GET",
            params={},
            data=None,
        )
        obj = ItemResultResponseBody(**response.json())
        obj.auth = self.auth
        return obj

    def get_items(
        self,
        collectionId: str = None,
        createdByUserId: str = None,
        id: str = None,
        limit: int = None,
        name: str = None,
        next: str = None,
        notCreatedByUserId: str = None,
        notOwnerId: str = None,
        ownerId: str = None,
        prev: str = None,
        query: str = None,
        resourceId: str = None,
        resourceIds: str = None,
        resourceLink: str = None,
        resourceSubType: str = None,
        resourceType: Literal[
            "app",
            "collection",
            "qlikview",
            "insight",
            "qvapp",
            "genericlink",
            "sharingservicetask",
            "note",
            "dataasset",
            "dataset",
            "automation",
            "automl-experiment",
            "automl-deployment",
        ] = None,
        shared: bool = None,
        sort: Literal[
            "+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"
        ] = None,
        spaceId: str = None,
        noActions: bool = False,
        max_items: int = 10,
    ) -> ListableResource[ItemResultResponseBody]:
        """
        Retrieves items that the user has access to.
        Finds and returns items that the user has access to.


        Parameters
        ----------
        collectionId: str = None
          The collection's unique identifier.
        createdByUserId: str = None
          User's unique identifier.
        id: str = None
          The item's unique identifier.
        limit: int = None
          The maximum number of resources to return for a request. The limit must be an integer between 1 and 100 (inclusive).
        name: str = None
          The case-insensitive string used to search for a resource by name.
        next: str = None
          The cursor to the next page of resources. Provide either the next or prev cursor, but not both.
        notCreatedByUserId: str = None
          User's unique identifier.
        notOwnerId: str = None
          Owner identifier.
        ownerId: str = None
          Owner identifier.
        prev: str = None
          The cursor to the previous page of resources. Provide either the next or prev cursor, but not both.
        query: str = None
          The case-insensitive string used to search for a resource by name or description.
        resourceId: str = None
          The case-sensitive string used to search for an item by resourceId. If resourceId is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
        resourceIds: str = None
          The case-sensitive strings used to search for an item by resourceIds. The maximum number of resourceIds it supports is 100. If resourceIds is provided, then resourceType must be provided. For example '?resourceIds=appId1,appId2'
        resourceLink: str = None
          The case-sensitive string used to search for an item by resourceLink. If resourceLink is provided, then resourceType must be provided. Provide either the resourceId or resourceLink, but not both.
        resourceSubType: str = None
          the case-sensitive string used to filter items by resourceSubType(s). For example '?resourceSubType=chart-monitoring,qix-df,qvd'
        resourceType: Literal["app", "collection", "qlikview", "insight", "qvapp", "genericlink", "sharingservicetask", "note", "dataasset", "dataset", "automation", "automl-experiment", "automl-deployment"] = None
          The case-sensitive string used to filter items by resourceType(s). For example '?resourceType=app,qvapp'
        shared: bool = None
          Whether or not to return items in a shared space.
        sort: Literal["+createdAt", "-createdAt", "+name", "-name", "+updatedAt", "-updatedAt"] = None
          The property of a resource to sort on (default sort is +createdAt). The supported properties are createdAt, updatedAt, and name. A property must be prefixed by + or   - to indicate ascending or descending sort order respectively.
        spaceId: str = None
          The space's unique identifier (supports \'personal\' as spaceId).
        noActions: bool = False
          If set to true, the user's available actions for each item will not be evaluated meaning the actions-array will be omitted from the response (reduces response time).

        """
        query_params = {}
        if collectionId is not None:
            query_params["collectionId"] = collectionId
        if createdByUserId is not None:
            query_params["createdByUserId"] = createdByUserId
        if id is not None:
            query_params["id"] = id
        if limit is not None:
            query_params["limit"] = limit
        if name is not None:
            query_params["name"] = name
        if next is not None:
            query_params["next"] = next
        if notCreatedByUserId is not None:
            query_params["notCreatedByUserId"] = notCreatedByUserId
        if notOwnerId is not None:
            query_params["notOwnerId"] = notOwnerId
        if ownerId is not None:
            query_params["ownerId"] = ownerId
        if prev is not None:
            query_params["prev"] = prev
        if query is not None:
            query_params["query"] = query
        if resourceId is not None:
            query_params["resourceId"] = resourceId
        if resourceIds is not None:
            query_params["resourceIds"] = resourceIds
        if resourceLink is not None:
            query_params["resourceLink"] = resourceLink
        if resourceSubType is not None:
            query_params["resourceSubType"] = resourceSubType
        if resourceType is not None:
            query_params["resourceType"] = resourceType
        if shared is not None:
            query_params["shared"] = shared
        if sort is not None:
            query_params["sort"] = sort
        if spaceId is not None:
            query_params["spaceId"] = spaceId
        if noActions is not None:
            query_params["noActions"] = noActions
        response = self.auth.rest(
            path="/items",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=ItemResultResponseBody,
            auth=self.auth,
            path="/items",
            max_items=max_items,
            query_params=query_params,
        )



# This is spectacularly generated code by spectacular based on
# Qlik Cloud Services APIs

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Literal

from ..auth import Auth, Config
from ..listable import ListableResource


@dataclass
class IDP:
    """

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPArray:
    """

    Attributes
    ----------
    data: list[IDP]
      An array of IdPs.
    links: Links
      Contains pagination links.
    """

    data: list[IDP] = None
    links: Links = None

    def __init__(self_, **kvargs):

        if "data" in kvargs and kvargs["data"] is not None:
            if all(
                f"list[{type(e).__name__}]" == IDPArray.__annotations__["data"]
                for e in kvargs["data"]
            ):
                self_.data = kvargs["data"]
            else:
                self_.data = [IDP(**e) for e in kvargs["data"]]
        if "links" in kvargs and kvargs["links"] is not None:
            if type(kvargs["links"]).__name__ == IDPArray.__annotations__["links"]:
                self_.links = kvargs["links"]
            else:
                self_.links = Links(**kvargs["links"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPMeta:
    """

    Attributes
    ----------
    upgradeSubscriptionLink: str
      A link to direct you to where you can upgrade your trial or manage your subscriptions. Only available if the default identity provider is used (no custom interactive identity providers are active).
    userPortalLink: str
      A link to direct you to where you can manage your Qlik account. Only available if the default identity provider is used (no custom interactive identity providers are active).
    """

    upgradeSubscriptionLink: str = None
    userPortalLink: str = None

    def __init__(self_, **kvargs):

        if (
            "upgradeSubscriptionLink" in kvargs
            and kvargs["upgradeSubscriptionLink"] is not None
        ):
            self_.upgradeSubscriptionLink = kvargs["upgradeSubscriptionLink"]
        if "userPortalLink" in kvargs and kvargs["userPortalLink"] is not None:
            self_.userPortalLink = kvargs["userPortalLink"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPPatchSchema:
    """

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPPostSchema:
    """

    Attributes
    ----------
    """

    def __init__(self_, **kvargs):
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPsStatus:
    """

    Attributes
    ----------
    active_interactive_idps_count: float
      The number of active interactive IdPs.
    idps_metadata: list[IDPsStatusIdpsMetadata]
      A list of IdP metadata.
    """

    active_interactive_idps_count: float = None
    idps_metadata: list[IDPsStatusIdpsMetadata] = None

    def __init__(self_, **kvargs):

        if (
            "active_interactive_idps_count" in kvargs
            and kvargs["active_interactive_idps_count"] is not None
        ):
            self_.active_interactive_idps_count = kvargs[
                "active_interactive_idps_count"
            ]
        if "idps_metadata" in kvargs and kvargs["idps_metadata"] is not None:
            if all(
                f"list[{type(e).__name__}]"
                == IDPsStatus.__annotations__["idps_metadata"]
                for e in kvargs["idps_metadata"]
            ):
                self_.idps_metadata = kvargs["idps_metadata"]
            else:
                self_.idps_metadata = [
                    IDPsStatusIdpsMetadata(**e) for e in kvargs["idps_metadata"]
                ]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class IDPsStatusIdpsMetadata:
    """

    Attributes
    ----------
    active: bool
      Indicates whether the IdP is available for use.
    interactive: bool
      Indicates whether the IdP is meant for interactive login.
    provider: Literal["auth0", "okta", "qlik", "generic", "salesforce", "keycloak", "adfs", "external", "azureAD"]
      The identity provider to be used. If protocol is `OIDC`, the valid values are `auth0`, `okta`, `generic`, `salesforce`, `keycloak`, `adfs`, and `azureAD`. If protocol is `jwtAuth`, the valid value is `external`.
    """

    active: bool = None
    interactive: bool = None
    provider: Literal[
        "auth0",
        "okta",
        "qlik",
        "generic",
        "salesforce",
        "keycloak",
        "adfs",
        "external",
        "azureAD",
    ] = None

    def __init__(self_, **kvargs):

        if "active" in kvargs and kvargs["active"] is not None:
            self_.active = kvargs["active"]
        if "interactive" in kvargs and kvargs["interactive"] is not None:
            self_.interactive = kvargs["interactive"]
        if "provider" in kvargs and kvargs["provider"] is not None:
            self_.provider = kvargs["provider"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class Links:
    """
    Contains pagination links.

    Attributes
    ----------
    next: LinksNext
    prev: LinksPrev
    self: LinksSelf
    """

    next: LinksNext = None
    prev: LinksPrev = None
    self: LinksSelf = None

    def __init__(self_, **kvargs):

        if "next" in kvargs and kvargs["next"] is not None:
            if type(kvargs["next"]).__name__ == Links.__annotations__["next"]:
                self_.next = kvargs["next"]
            else:
                self_.next = LinksNext(**kvargs["next"])
        if "prev" in kvargs and kvargs["prev"] is not None:
            if type(kvargs["prev"]).__name__ == Links.__annotations__["prev"]:
                self_.prev = kvargs["prev"]
            else:
                self_.prev = LinksPrev(**kvargs["prev"])
        if "self" in kvargs and kvargs["self"] is not None:
            if type(kvargs["self"]).__name__ == Links.__annotations__["self"]:
                self_.self = kvargs["self"]
            else:
                self_.self = LinksSelf(**kvargs["self"])
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LinksNext:
    """

    Attributes
    ----------
    href: str
      Link to the next page of items.
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LinksPrev:
    """

    Attributes
    ----------
    href: str
      Link to the previous page of items.
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


@dataclass
class LinksSelf:
    """

    Attributes
    ----------
    href: str
      Link to the current page of items.
    """

    href: str = None

    def __init__(self_, **kvargs):

        if "href" in kvargs and kvargs["href"] is not None:
            self_.href = kvargs["href"]
        for k0, v in kvargs.items():
            k = k0.replace("-", "_")
            if k not in getattr(self_, "__annotations__", {}):
                self_.__setattr__(k, v)


class IdentityProviders:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.auth = Auth(config)

    def get_well_known_metadata_json(self) -> object:
        """
        Returns identity providers' metadata
        This endpoint retrieves identity providers' metadata.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/identity-providers/.well-known/metadata.json",
            method="GET",
            params={},
            data=None,
        )
        return response.json()

    def get_me_meta(self) -> IDPMeta:
        """
        Returns the active interactive IdP metadata
        This endpoint retrieves IdP metadata.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/identity-providers/me/meta",
            method="GET",
            params={},
            data=None,
        )
        obj = IDPMeta(**response.json())
        obj.auth = self.auth
        return obj

    def get_status(self) -> IDPsStatus:
        """
        Returns the current status of IdP configurations
        This endpoint retrieves the status of IdP configurations. Requires TenantAdmin role.

        Parameters
        ----------
        """
        response = self.auth.rest(
            path="/identity-providers/status",
            method="GET",
            params={},
            data=None,
        )
        obj = IDPsStatus(**response.json())
        obj.auth = self.auth
        return obj

    def delete(self, id: str) -> None:
        """
        Deletes the IdP with the specified ID
        This endpoint deletes an identity provider from the service. It returns a valid 204 when the IdP is deleted. Only a user with the role of TenantAdmin and tenant access can delete an associated IdP. Edge-auth service can also delete.

        Parameters
        ----------
        id: str
          The identity provider ID.
        """
        self.auth.rest(
            path="/identity-providers/{id}".replace("{id}", id),
            method="DELETE",
            params={},
            data=None,
        )

    def get(self, id: str) -> IDP:
        """
        Returns the IdP with the specified ID
        This endpoint is used to retrieve an identity provider from the service. It returns a valid 200 OK response when the IdP exists and the user (TenantAdmin) or service (edge-auth) is authorized to view the contents. Additionally, returns a header "QLIK-IDP-POPTS" (A unique string representing a hash of the current configuration being tested). It returns a 404 Not Found if the criteria is not met.

        Parameters
        ----------
        id: str
          The identity provider ID.
        """
        response = self.auth.rest(
            path="/identity-providers/{id}".replace("{id}", id),
            method="GET",
            params={},
            data=None,
        )
        obj = IDP(**response.json())
        obj.auth = self.auth
        return obj

    def patch(self, id: str, data: IDPPatchSchema = None) -> None:
        """
        Updates the IdP with the specified ID
        This endpoint patches an identity provider from the service. It returns a valid 204 when the IdP is patched. Only an edge-auth service request or a user with the role of TenantAdmin can patch an associated IdP. Partial failure is treated as complete failure and returns an error.

        Parameters
        ----------
        id: str
          The identity provider ID.
        QLIK-IDP-POPTS-MATCH: str = None
          A unique string representing a hash that should map to an IdP's hash representation of the current configuration being tested.
        data: IDPPatchSchema = None
          Attributes that the user wants to patially update for an identity provider resource.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        self.auth.rest(
            path="/identity-providers/{id}".replace("{id}", id),
            method="PATCH",
            params={},
            data=data,
        )

    def get_identity_providers(
        self,
        active: bool = None,
        limit: float = 20,
        next: str = None,
        prev: str = None,
        max_items: int = 20,
    ) -> ListableResource[IDP]:
        """
        Retrieves one or more IdPs for a specified tenantId.
        This endpoint retrieves one or more identity providers from the service. The tenantID in the JWT will be used to fetch the identity provider.

        Parameters
        ----------
        active: bool = None
          If provided, filters the results by the active field.
        limit: float = 20
          The number of IdP entries to retrieve.
        next: str = None
          The next page cursor.
        prev: str = None
          The previous page cursor.
        """
        query_params = {}
        if active is not None:
            query_params["active"] = active
        if limit is not None:
            query_params["limit"] = limit
        if next is not None:
            query_params["next"] = next
        if prev is not None:
            query_params["prev"] = prev
        response = self.auth.rest(
            path="/identity-providers",
            method="GET",
            params=query_params,
            data=None,
        )
        return ListableResource(
            response=response.json(),
            cls=IDP,
            auth=self.auth,
            path="/identity-providers",
            max_items=max_items,
            query_params=query_params,
        )

    def create(self, data: IDPPostSchema = None) -> IDP:
        """
        Creates a new IdP
        This endpoint creates an identity provider resource. It returns a 201 Created when creation is successful with a header "QLIK-IDP-POPTS" (A unique string representing a hash of the current configuration being tested), returns a 403 Forbidden for a non TenantAdmin user JWT or if the tenantID in the JWT does not match with any of the tenantIDs in the payload. An IdP can be created with Pending Options or options depending whether the IdP is interactive or not.

        Parameters
        ----------
        data: IDPPostSchema = None
          Attributes that the user wants to set for a new identity provider resource.
        """
        if data is not None:
            try:
                data = asdict(data)
            except:
                data = data
        response = self.auth.rest(
            path="/identity-providers",
            method="POST",
            params={},
            data=data,
        )
        obj = IDP(**response.json())
        obj.auth = self.auth
        return obj
